//==================================================================
//
//	File name	:	treeview.js
//
//	Summary		:	TreeView class
//
//	Ceated		:	2018/12/15
//	Written by	:	Yuji Iihara
//	Copyright	:	Softinc.
//
//==================================================================
"use strict";

(function (treeview) {
	//==========================================================
	//	Global
	//==========================================================
	// Const
	treeview.STYLE_ICON = 0;
	treeview.STYLE_IMAGE = 1;
	treeview.STYLE_EDITABLE = 2;
	
	
	//==========================================================
	//
	//	Class		:	Node
	//
	//	Summary		:	Class of treeview node
	//
	//==========================================================
	treeview.Node = class
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sLabel,			// * Label text[String]
					sValue,			// * Node value[String]
					iStyle,			// * Node style[int]
					sContents,		// Icon name or image URL[String]
					fpClick,		// Function on click[function]
					fpDblClick,		// Function on double click[function]
					bDraggable,		// Draggable or not[boolean]
					fpDragstart,	// Function on start dragging[function]
					bDroppable,		// Droppable or not[boolean]
					iDropType,		// Drop type[int]
					fpDrop)			// Function on dropped[function]
		{
			// Check parameters
			if(typeof sLabel != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, ""),
											"treeview.js", "jaswin.treeview.Node.constructor");
			
			if(typeof sValue != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, ""),
											"treeview.js", "jaswin.treeview.Node.constructor");
			
			if(typeof iStyle != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, ""),
											"treeview.js", "jaswin.treeview.Node.constructor");
			if(iStyle & treeview.STYLE_EDITABLE)
			{
				if(!jaswin.textbox)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_INCLUDE, "textbox.js"),
												"treeview.js", "jaswin.treeview.Node.constructor");
			}
			
			if(sContents)
			{
				if(typeof sContents != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_CONTENTS, ""),
												"treeview.js", "jaswin.treeview.Node.constructor");
			}
			
			if(fpClick)
			{
				if(typeof fpClick != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CLICK, ""),
												"treeview.js", "jaswin.treeview.Node.constructor");
			}
			
			if(fpDblClick)
			{
				if(typeof fpDblClick != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CLICK, ""),
												"treeview.js", "jaswin.treeview.Node.constructor");
			}
			
			if(bDraggable)
			{
				if(typeof bDraggable != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DRAGGABLE, ""),
												"treeview.js", "jaswin.treeview.Node.constructor");
				if(fpDragstart)
				{
					if(typeof fpDragstart != "function")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_ON_DRAGSTART, ""),
													"treeview.js", "jaswin.treeview.Node.constructor");
				}
			}
			
			if(bDroppable)
			{
				if(typeof bDroppable != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DROPPABLE, ""),
												"treeview.js", "jaswin.treeview.Node.constructor");
				
				if(typeof iDropType != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DROP_TYPE, ""),
												"treeview.js", "jaswin.treeview.Node.constructor");
				if((iDropType != jaswin.DROP_TEXT)
				&& (iDropType != jaswin.DROP_URL)
				&& (iDropType != jaswin.DROP_HTML)
				&& (iDropType != jaswin.DROP_FILE))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_DROP_TYPE, ""),
												"treeview.js", "jaswin.treeview.Node.constructor");
				
				if(fpDrop)
				{
					if(typeof fpDrop != "function")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_ON_DROP, ""),
													"treeview.js", "jaswin.treeview.Node.constructor");
				}
			}
			
			// Set properties
			this.value = sValue;				// Node value[String]
			this.style = iStyle;				// Node style[int]
			this.onClick = fpClick;				// Function on click[function]
			this.onDblClick = fpDblClick;		// Function on double click[function]
			this.element = null;				// This element[DOM Element]
			this.nodeItem = null;				// Node item element[DOM Element]
			this.label = null;					// Label element[DOM Element]
			this.box = null;					// Icon box of open or close[DOM Element]
			this.openClose = null;				// Icon of open or close[DOM Element]
			this.icon = null;					// Icon element[DOM Element]
			this.opened = false;				// Open or not[boolean]
			this.selected = false;				// Selected or not[boolean]
			this.container = null;				// Container element[DOM Element]
			this.childNodes = new Array();		// Child nodes[Array of Node]
			this.parentNode = null;				// ParentNode[Node]
			this.treeView = null;				// Treeview[TreeView class]
			this.editable = false;				// Label is editable or not[boolean]
			this.draggable = bDraggable;		// Draggable or not[boolean]
			this.onDragstart = fpDragstart;		// Function on dragstart[function(Argument:ListItem class, event)]
			this.droppable = false;				// Droppable or not[boolean]
			this.dropType = iDropType;			// Drop type[int]
			this.onDrop = null;					// Function on drop[function(Argument:Node class, e)]
			
			if(bDroppable)
			{
				this.droppable = bDroppable;
				this.onDrop = fpDrop;
			}
			
			// Create element
			this.element = document.createElement("li");
			this.nodeItem = document.createElement("div");
			this.nodeItem.className = "tree-node-item";
			this.element.appendChild(this.nodeItem);
			
			// Box
			this.box = document.createElement("div");
			this.box.className = "tree-node-box";
			this.openClose = document.createElement("span");
			this.openClose.className = "tree-icon-close";
			this.box.appendChild(this.openClose);
			this.nodeItem.appendChild(this.box);
			
			// Icon
			this.icon = document.createElement("div");
			this.icon.className = "tree-node-icon";
			
			if(!sContents)
			{
				let oSpan = document.createElement("span");
				oSpan.className = jaswin.ICON_FOLDER;
				this.icon.appendChild(oSpan);
			}
			else
			{
				if(iStyle & treeview.STYLE_IMAGE)
				{
					let oImg = document.createElement("img");
					oImg.src = sContents;
					oImg.width = 16;
					oImg.height = 16;
					this.icon.appendChild(oImg);
				}
				else
				{
					let oSpan = document.createElement("span");
					oSpan.className = sContents;
					this.icon.appendChild(oSpan);
				}
			}
			this.nodeItem.appendChild(this.icon);
			if(iStyle & treeview.STYLE_EDITABLE)
				this.editable = true;
			
			// Label
			this.label = document.createElement("span");
			this.label.className = "tree-node-label";
			this.label.innerHTML = sLabel;
			this.nodeItem.appendChild(this.label);
			
			let oDiv = document.createElement("div");
			oDiv.style.clear = "both";
			this.nodeItem.appendChild(oDiv);
		}
		
		//==================================================
		//	Set parent treeview
		//==================================================
		setTreeView(oTreeView)
		{
			if((typeof oTreeView != "object")
			|| (!oTreeView instanceof treeview.TreeView))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_TREEVIEW,
											"treeview.js", "jaswin.treeview.Node.setTreeView");
			
			this.treeView = oTreeView;
			if(this.treeView)
			{
				if(this.treeView.enabled)
					this.enable();
				else
					this.disable();
				
				let iMinWidth = this.label.offsetWidth + 10 + 16 + 2;	// 10:Box 16:Icon 2:border
				this.nodeItem.style.minWidth = iMinWidth + "px";
			}
			
			if(this.container)
			{
				let iCounter;
				for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
					this.childNodes[iCounter].setTreeView(oTreeView);
			}
		}
		
		//==================================================
		//	Enable this tree node
		//==================================================
		enable()
		{
			if(!this.treeView)
				return;
			
			this.nodeItem.className = "tree-node-item";
			this.icon.style.cursor = "pointer";
			this.label.style.cursor = "pointer";
			
			if(!this.boxClickEvent)
			{
				this.box.addEventListener("click", this.boxClickEvent = (e) => {
					this.clickBox();
				}, false);
			}
			if(!this.iconClickEvent)
			{
				this.icon.addEventListener("click", this.iconClickEvent = (e) => {
					this.treeView.clickNode(this, false);
				}, false);
			}
			if(!this.iconDblClickEvent)
			{
				this.icon.addEventListener("dblclick", this.iconDblClickEvent = (e) => {
					this.dblClick();
				}, false);
			}
			if(!this.iconContextMenuEvent)
			{
				this.icon.addEventListener("contextmenu", this.iconContextMenuEvent = (e) => {
					this.treeView.contextMenuNode(e, this);
				}, false);
			}
			if(!this.labelClickEvent)
			{
				this.label.addEventListener("click", this.labelClickEvent = (e) => {
					this.treeView.clickNode(this, true);
				}, false);
			}
			if(!this.labelDblClickEvent)
			{
				this.label.addEventListener("dblclick", this.labelDblClickEvent = (e) => {
					this.dblClick();
				}, false);
			}
			if(!this.labelContextMenuEvent)
			{
				this.label.addEventListener("contextmenu", this.labelContextMenuEvent = (e) => {
					this.treeView.contextMenuNode(e, this);
				}, false);
			}
			if(this.draggable)
			{
				if(!this.dragStartEvent)
				{
					this.nodeItem.draggable = true;
					this.nodeItem.addEventListener("dragstart", this.dragStartEvent = (e) => {
						this.onDragstart(this, e);
					}, false);
				}
			}
			if(this.droppable)
			{
				if(!this.dragOverEvent)
				{
					this.nodeItem.addEventListener("dragover", this.dragOverEvent = (e) => {
						let iCounter;
						let bFound = false;
						
						for(iCounter = 0; iCounter < e.dataTransfer.types.length; iCounter++)
						{
							if((this.dropType == jaswin.DROP_TEXT)
							&& (e.dataTransfer.types[iCounter] == "text/plain"))
								bFound = true;
							else if((this.dropType == jaswin.DROP_URL)
							&& (e.dataTransfer.types[iCounter] == "text/uri-list"))
								bFound = true;
							else if((this.dropType == jaswin.DROP_HTML)
							&& (e.dataTransfer.types[iCounter] == "text/html"))
								bFound = true;
							else if((this.dropType == jaswin.DROP_FILE)
							&& (e.dataTransfer.types[iCounter] == "Files"))
								bFound = true;
						}
						if(!bFound)
							return;
						e.dataTransfer.dropEffect = "copy";
						e.stopPropagation();
						e.preventDefault();
					}, false);
				}
				if(!this.dropEvent)
				{
					this.nodeItem.addEventListener("drop", this.dropEvent = (e) => {
						e.stopPropagation();
						e.preventDefault();
						if(this.onDrop)
							this.onDrop(this, e);
					}, false);
				}
			}
			
			if(this.container)
			{
				let iCounter;
				for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
					this.childNodes[iCounter].enable();
			}
		}
		
		//==================================================
		//	Disable this tree node
		//==================================================
		disable()
		{
			if(!this.treeView)
				return;
			
			this.nodeItem.className = "tree-node-item-disabled";
			this.icon.style.cursor = "default";
			this.label.style.cursor = "default";
			
			if(this.boxClickEvent)
			{
				this.box.removeEventListener("click", this.boxClickEvent, false);
				this.boxClickEvent = null;
			}
			if(this.iconClickEvent)
			{
				this.icon.removeEventListener("click", this.iconClickEvent, false);
				this.iconClickEvent = null;
			}
			if(this.iconDblClickEvent)
			{
				this.icon.removeEventListener("dblclick", this.iconDblClickEvent, false);
				this.iconDblClickEvent = null;
			}
			if(this.iconContextMenuEvent)
			{
				this.icon.removeEventListener("contextmenu", this.iconContextMenuEvent, false);
				this.iconContextMenuEvent = null;
			}
			if(this.labelClickEvent)
			{
				this.label.removeEventListener("click", this.labelClickEvent, false);
				this.labelClickEvent = null;
			}
			if(this.labelDblClickEvent)
			{
				this.label.removeEventListener("dblclick", this.labelDblClickEvent, false);
				this.labelDblClickEvent = null;
			}
			if(this.labelContextMenuEvent)
			{
				this.label.removeEventListener("contextmenu", this.labelContextMenuEvent, false);
				this.labelContextMenuEvent = null;
			}
			if(this.draggable)
			{
				this.nodeItem.draggable = false;
				if(this.dragStartEvent)
				{
					this.nodeItem.removeEventListener("dragstart", this.dragStartEvent, false);
					this.dragStartEvent = null;
				}
			}
			if(this.dragOverEvent)
			{
				this.nodeItem.removeEventListener("dragover", this.dragOverEvent, false);
				this.dragOverEvent = null;
			}
			if(this.dropEvent)
			{
				this.nodeItem.removeEventListener("drop", this.dropEvent, false);
				this.dropEvent = null;
			}
			
			if(this.container)
			{
				let iCounter;
				for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
					this.childNodes[iCounter].disable();
			}
		}
		
		//==================================================
		//	Add child node to this node
		//==================================================
		addNode(oChildNode)
		{
			if((typeof oChildNode != "object")
			|| (!oChildNode instanceof treeview.Node))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_NODE, ""),
											"treeview.js", "jaswin.treeview.Node.addNode");
			
			if(!this.container)
			{
				this.container = document.createElement("ul");
				this.container.className = "tree-list";
				this.container.style.display = "none";
				this.element.appendChild(this.container);
			}
			
			this.container.appendChild(oChildNode.element);
			this.childNodes[this.childNodes.length] = oChildNode;
			oChildNode.parentNode = this;
			oChildNode.setTreeView(this.treeView);
		}
		
		//==================================================
		//	Remove this node
		//==================================================
		removeNode()
		{
			let oParentNode = this.parentNode;
			let iCounter;
			let iIndex = -1;
			
			if(!oParentNode)
				return;
			
			for(iCounter = 0; iCounter < oParentNode.childNodes.length; iCounter++)
			{
				if(oParentNode.childNodes[iCounter] == this)
				{
					iIndex = iCounter;
					break;
				}
			}
			if(iIndex == -1)
				return;
			
			oParentNode.container.removeChild(this.element);
			oParentNode.childNodes.splice(iIndex, 1);
			if((oParentNode.childNodes.length == 0)
			&& (oParentNode instanceof treeview.Node))
			{
				oParentNode.element.removeChild(oParentNode.container);
				oParentNode.container = null;
			}
		}
		
		//==================================================
		//	Set label text
		//==================================================
		setLabel(sText)
		{
			if(typeof sText != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LABEL, this.id),
											"treeview.js", "jaswin.treeview.Node.setLabel");
			
			this.label.innerHTML = sText;
			if(this.treeView)
			{
				let iMinWidth = this.label.offsetWidth + 10 + 16 + 2;	// 10:Box 16:Icon 2:border
				this.nodeItem.style.minWidth = iMinWidth + "px";
			}
		}
		
		//==================================================
		//	Set this node hover
		//==================================================
		setHover()
		{
			if(this.childNodes.length == 0)
				return true;
			
			this.openClose.style.visibility = "visible";
			if(this.opened)
			{
				let iCounter;
				for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
					this.childNodes[iCounter].setHover();
			}
		}
		
		//==================================================
		//	Set this node mouse out
		//==================================================
		setOut()
		{
			if(this.childNodes.length == 0)
				return true;
			
			this.openClose.style.visibility = "hidden";
			if(this.opened)
			{
				let iCounter;
				for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
					this.childNodes[iCounter].setOut();
			}
		}
		
		//==================================================
		//	Click event of this node box
		//==================================================
		clickBox()
		{
			if(this.opened)
				this.close();
			else
				this.open();
		}
		
		//==================================================
		//	Open this node
		//==================================================
		open()
		{
			if(!this.container)
				return;
			
			let iCounter;
			let iMinWidth;
			
			this.openClose.className = "tree-icon-open";
			this.container.style.display = "block";
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
			{
				iMinWidth = this.childNodes[iCounter].label.offsetWidth + 10 + 16 + 2;	// 10:Box 16:Icon 2:border
				this.childNodes[iCounter].nodeItem.style.minWidth = iMinWidth + "px";
			}
			this.opened = true;
		}
		
		//==================================================
		//	Close this node
		//==================================================
		close()
		{
			if(!this.container)
				return;
			
			this.openClose.className = "tree-icon-close";
			this.container.style.display = "none";
			this.opened = false;
			return true;
		}
		
		//==================================================
		//	Double click event of this node
		//==================================================
		dblClick()
		{
			if(this.onDblClick)
			{
				this.onDblClick(this);
				return;
			}
			
			if(this.childNodes.length > 0)
				this.clickBox();
		}
		
		//==================================================
		//	Select a node
		//==================================================
		selectNode(oTargetNode)
		{
			if((typeof oTargetNode != "object")
			|| (!oTargetNode instanceof treeview.Node))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_NODE, ""),
											"treeview.js", "jaswin.treeview.Node.selectNode");
			
			if(oTargetNode == this)
			{
				this.selected = true;
				this.nodeItem.className = "tree-node-item-focus";
				
				let oBound = this.treeView.element.getBoundingClientRect();
				let iTop = this.treeView.element.scrollTop;
				let iBottom = this.treeView.element.scrollTop + oBound.height;
				
				if(this.element.offsetTop < iTop)
					this.treeView.element.scrollTop = this.element.offsetTop;
				else if(this.element.offsetTop + 24 > iBottom)
					this.treeView.element.scrollTop += this.element.offsetTop + 24 - iBottom;
				
				if(this.onClick)
					this.onClick(this);
			}
			else
			{
				this.selected = false;
				this.nodeItem.className = "tree-node-item";
			}
			
			if(this.container)
			{
				let iCounter;
				for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
					this.childNodes[iCounter].selectNode(oTargetNode);
			}
			
			return true;
		}
		
		//==================================================
		//	Get selected node
		//==================================================
		getSelectedNode()
		{
			if(this.selected)
				return this;
			
			let iCounter;
			let oSelected = null;
			
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
			{
				oSelected = this.childNodes[iCounter].getSelectedNode();
				if(oSelected)
					return oSelected;
			}
			
			return null;
		}
		
		//==================================================
		//	Find next node
		//==================================================
		findNext(oNode)
		{
			if((typeof oNode != "object")
			|| (!oNode instanceof treeview.Node))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_NODE, ""),
											"treeview.js", "jaswin.treeview.Node.findNext");
			
			if(oNode == this)
			{
				if(this.opened)
					return this.childNodes[0];
				
				return this.parentNode.findNext(this);
			}
			
			let iCounter;
			let iIndex = -1;
			
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
			{
				if(this.childNodes[iCounter] == oNode)
				{
					iIndex = iCounter;
					break;
				}
			}
			if((iIndex == -1) || (iIndex == this.childNodes.length - 1))
				return this.parentNode.findNext(this);
			
			return this.childNodes[iIndex + 1];
		}
		
		//==================================================
		//	Find previous node
		//==================================================
		findPrevious(oNode)
		{
			if((typeof oNode != "object")
			|| (!oNode instanceof treeview.Node))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_NODE, ""),
											"treeview.js", "jaswin.treeview.Node.findPrevious");
			
			let iCounter;
			let iIndex = -1;
			
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
			{
				if(this.childNodes[iCounter] == oNode)
				{
					iIndex = iCounter;
					break;
				}
			}
			
			if(iIndex <= 0)
				return this;
			
			if(this.childNodes[iIndex - 1].opened)
			{
				let iLength = this.childNodes[iIndex - 1].childNodes.length;
				
				return this.childNodes[iIndex - 1].childNodes[iLength - 1];
			}
			
			return this.childNodes[iIndex - 1];
		}
		
		//==================================================
		//	Show label text box
		//==================================================
		showLabelTextBox()
		{
			if(!this.treeView.labelTextBox)
				return;
			
			let iLeft;
			let iTop;
			let iWidth;
			
			if(this.treeView.labelTextBox.style == jaswin.textbox.STYLE_FLAT)
			{
				iLeft = parseInt(this.nodeItem.offsetLeft) + 32;
				iTop = parseInt(this.nodeItem.offsetTop) + 2;
				iWidth = parseInt(this.nodeItem.offsetWidth) - 35;
			}
			else
			{
				iLeft = parseInt(this.nodeItem.offsetLeft) + 31;
				iTop = parseInt(this.nodeItem.offsetTop) + 1;
				iWidth = parseInt(this.nodeItem.offsetWidth) - 36;
			}
			this.treeView.labelTextBox.element.style.left = iLeft + "px";
			this.treeView.labelTextBox.element.style.top = iTop + "px";
			this.treeView.labelTextBox.element.style.width = iWidth + "px";
			this.treeView.labelTextBox.setValue(this.label.innerHTML);
			this.treeView.labelTextBox.show();
			this.treeView.labelTextBox.element.focus();
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	TreeView
	//
	//	Summary		:	Class of treeview
	//
	//==========================================================
	treeview.TreeView = class extends jaswin.Container
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[string]
					oRect,			// * Size and position[ExpandableRect]
					bEnabled,		// * Enabled or not[boolean]
					iTabIndex)		// Tab index[int]
		{
			// Base class
			super(sID, oRect, bEnabled, iTabIndex);
			
			// Set properties
			this.childNodes = new Array();	// Child nodes[Array of Node]
			this.labelTextBox = null;		// TextBox for editing node label[TextBox]
			
			// Create element
			this.element = document.createElement("div");
			this.element.className = "treeview";
			this.focusElement = this.element;
			
			// Horizontal rect
			if(this.rect.anchorX == jaswin.ANCHOR_BOTH)
			{
				this.element.style.left = this.rect.left + "px";
				this.element.style.right = this.rect.right + "px";
			}
			else
			{
				if(this.rect.width < this.minWidth)
					this.rect.width = this.minWidth;
				
				if(this.rect.anchorX == jaswin.ANCHOR_LEFT)
					this.element.style.left = this.rect.left + "px";
				else if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
					this.element.style.right = this.rect.right + "px";
				this.element.style.width = this.rect.width - 2 + "px";		// 2:border
			}
			
			// Vertical rect
			if(this.rect.anchorY == jaswin.ANCHOR_BOTH)
			{
				this.element.style.top = this.rect.top + "px";
				this.element.style.bottom = this.rect.bottom + "px";
			}
			else
			{
				if(this.rect.anchorY == jaswin.ANCHOR_TOP)
					this.element.style.top = this.rect.top + "px";
				else if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
					this.element.style.bottom = this.rect.bottom + "px";
				this.element.style.height = this.rect.height - 2 + "px";	// 2:border
			}
			
			// Create container
			this.container = document.createElement("ul");
			this.container.className = "tree-list";
			this.element.appendChild(this.container);
			
			// Tab index
			if(iTabIndex)
			{
				if(iTabIndex > 0)
				{
					this.tabIndex = iTabIndex;
					this.focusElement.tabIndex = this.tabIndex;
				}
			}
			
			// Event binding
			if(this.enabled)
			{
				this.element.addEventListener("mousemove", this.mouseMoveEvent = (e) => {
					this.setHover();
				}, false);
				this.element.addEventListener("mouseout", this.mouseOutEvent = (e) => {
					this.setOut();
				}, false);
				this.element.addEventListener("keydown", this.keyEvent = (e) => {
					this.keyDown(e);
				}, false);
			}
		}
		
		//==================================================
		//	Delete DOM element
		//==================================================
		destroy()
		{
			if(this.labelTextBox)
				this.labelTextBox.destroy();
			
			super.destroy();
		}
		
		//==================================================
		//	Enable this treeview
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(this.enabled)
			{
				if(!this.mouseMoveEvent)
				{
					this.element.addEventListener("mousemove", this.mouseMoveEvent = (e) => {
						this.setHover();
					}, false);
				}
				if(!this.mouseOutEvent)
				{
					this.element.addEventListener("mouseout", this.mouseOutEvent = (e) => {
						this.setOut();
					}, false);
				}
				if(!this.keyEvent)
				{
					this.element.addEventListener("keydown", this.keyEvent = (e) => {
						this.keyDown(e);
					}, false);
				}
				
				let iCounter;
				for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
					this.childNodes[iCounter].enable();
			}
		}
		
		//==================================================
		//	Disable this treeview
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			if(this.mouseMoveEvent)
			{
				this.element.removeEventListener("mousemove", this.mouseMoveEvent, false);
				this.mouseMoveEvent = null;
			}
			if(this.mouseOutEvent)
			{
				this.element.removeEventListener("mouseout", this.mouseOutEvent, false);
				this.mouseOutEvent = null;
			}
			if(this.keyEvent)
			{
				this.element.removeEventListener("keydown", this.keyEvent, false);
				this.keyEvent = null;
			}
			
			let iCounter;
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
				this.childNodes[iCounter].disable();
		}
		
		//==================================================
		//	Add Node
		//==================================================
		addNode(oNode)
		{
			if((typeof oNode != "object")
			|| (!oNode instanceof treeview.Node))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_NODE, this.id),
											"treeview.js", "jaswin.treeview.TreeView.addNode");
			
			this.container.appendChild(oNode.element);
			oNode.parentNode = this;
			oNode.setTreeView(this);
			
			this.childNodes.push(oNode);
		}
		
		//==================================================
		//	Set this treeview hover
		//==================================================
		setHover()
		{
			let iCounter;
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
				this.childNodes[iCounter].setHover();
		}
		
		//==================================================
		//	Set this treeview mouse out
		//==================================================
		setOut()
		{
			let iCounter;
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
				this.childNodes[iCounter].setOut();
		}
		
		//==================================================
		//	Click event of node
		//==================================================
		clickNode(oNode, bLabel)
		{
			if(oNode.editable && this.labelTextBox && bLabel)
			{
				let oSelected = this.getSelectedNode();
				
				if(oNode == oSelected)
				{
					oNode.showLabelTextBox();
					return;
				}
			}
			
			let iCounter;
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
				this.childNodes[iCounter].selectNode(oNode);
		}
		
		//==================================================
		//	Set context menu to this treeview
		//==================================================
		setContextMenu(oMenu, fpContext)
		{
			super.setContextMenu(oMenu, fpContext);
			
			let oParentWnd = jaswin.findParentWindow(this);
			oParentWnd.container.appendChild(oMenu.element);
			
			oMenu.setParent(this, 0);
		}
		
		//==================================================
		//	Context menu event of node
		//==================================================
		contextMenuNode(e, oNode)
		{
			let iCounter;
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
				this.childNodes[iCounter].selectNode(oNode);
			
			if(this.contextMenu)
			{
				if(this.onContextMenu)
				{
					if(!this.onContextMenu(oNode))
					{
						jaswin.menu.menuCancel();
						return;
					}
				}
				this.contextMenu.showContext(e);
			}
		}
		
		//==================================================
		//	Set text box for editing node label
		//==================================================
		setLabelTextBox(oTextBox)
		{
			if(!this.parent)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_CONTROL_ADD, this.id),
											"treeview.js", "jaswin.treeview.TreeView.setLabelTextBox");
			
			if(!jaswin.textbox)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_INCLUDE, "textbox.js"),
											"treeview.js", "jaswin.treeview.TreeView.setLabelTextBox");
			
			if((typeof oTextBox != "object")
			|| (!oTextBox instanceof jaswin.textbox.TextBox))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_TEXTBOX, this.id),
											"treeview.js", "jaswin.treeview.TreeView.setLabelTextBox");
			this.labelTextBox = oTextBox;
			this.labelTextBox.hide();
			this.container.appendChild(this.labelTextBox.element);
			this.labelTextBox.setParent(this);
		}
		
		//==================================================
		//	Remove text box for editing node label
		//==================================================
		removeLabelTextBox()
		{
			if(this.labelTextBox)
			{
				this.labelTextBox.destroy();
				this.labelTextBox = null;
			}
		}
		
		//==================================================
		//	Select a node
		//==================================================
		selectNode(oNode)
		{
			if((typeof oNode != "object")
			|| (!oNode instanceof treeview.Node))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_NODE, this.id),
											"treeview.js", "jaswin.treeview.TreeView.selectNode");
			
			let iCounter;
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
				this.childNodes[iCounter].selectNode(oNode);
			
			if(this.labelTextBox)
				this.labelTextBox.hide();
		}
		
		//==================================================
		//	Get selected node
		//==================================================
		getSelectedNode()
		{
			if(this.childNodes.length == 0)
				return null;
			
			let oNode = null;
			let iCounter;
			
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
			{
				oNode = this.childNodes[iCounter].getSelectedNode();
				if(oNode)
					return oNode;
			}
			
			return oNode;
		}
		
		//==================================================
		//	Find next node
		//==================================================
		findNext(oNode)
		{
			if((typeof oNode != "object")
			|| (!oNode instanceof treeview.Node))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_NODE, this.id),
											"treeview.js", "jaswin.treeview.TreeView.findNext");
			
			let iCounter;
			let iIndex = -1;
			
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
			{
				if(this.childNodes[iCounter] == oNode)
				{
					iIndex = iCounter;
					break;
				}
			}
			
			if((iIndex == -1) || (iIndex == this.childNodes.length - 1))
				return null;
			
			return this.childNodes[iIndex + 1];
		}
		
		//==================================================
		//	Find previous node
		//==================================================
		findPrevious(oNode)
		{
			if((typeof oNode != "object")
			|| (!oNode instanceof treeview.Node))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_NODE, this.id),
											"treeview.js", "jaswin.treeview.TreeView.findPrevious");
			
			let iCounter;
			let iIndex = -1;
			
			for(iCounter = 0; iCounter < this.childNodes.length; iCounter++)
			{
				if(this.childNodes[iCounter] == oNode)
				{
					iIndex = iCounter;
					break;
				}
			}
			
			if(iIndex <= 0)
				return null;
			
			if(this.childNodes[iIndex - 1].opened)
			{
				let iLength = this.childNodes[iIndex - 1].childNodes.length;
				
				return this.childNodes[iIndex - 1].childNodes[iLength - 1];
			}
			
			return this.childNodes[iIndex - 1];
		}
		
		//==================================================
		//	Clear nodes
		//==================================================
		clear()
		{
			let iCounter;
			let iCnt = this.childNodes.length;
			
			for(iCounter = 0; iCounter < iCnt; iCounter++)
				this.childNodes[0].removeNode();
		}
		
		//==================================================
		//	Key down event of node box
		//==================================================
		keyDown(e)
		{
			if((this.contextMenu) && (this.contextMenu.isVisible()))
				return this.contextMenu.menuKey(e);
			
			if((this.labelTextBox) && (this.labelTextBox.isVisible()))
				return true;
			
			if(this.childNodes.length == 0)
				return true;
			
			let oNode = null;
			let oNext = null;
			let oPrev = null;
			
			if(e.keyCode == 40)
			{
				// down
				oNode = this.getSelectedNode();
				if(!oNode)
				{
					this.selectNode(this.childNodes[0]);
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				
				if(oNode.opened)
				{
					this.selectNode(oNode.childNodes[0]);
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				
				oNext = oNode.findNext(oNode);
				if(oNext)
				{
					this.selectNode(oNext);
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				else
					return true;
			}
			else if(e.keyCode == 38)
			{
				// up
				oNode = this.getSelectedNode();
				if(!oNode)
					return true;
				
				oPrev = oNode.parentNode.findPrevious(oNode);
				if(oPrev)
				{
					this.selectNode(oPrev);
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				else
					return true;
			}
			else if(e.keyCode == 39)
			{
				// right
				oNode = this.getSelectedNode();
				if(!oNode)
					return true;
				if(oNode.childNodes.length == 0)
					return true;
				
				if(oNode.opened)
				{
					this.selectNode(oNode.childNodes[0]);
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				else
				{
					oNode.clickBox();
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
			}
			else if(e.keyCode == 37)
			{
				// left
				oNode = this.getSelectedNode();
				if(!oNode)
					return true;
				
				if(oNode.opened)
				{
					oNode.clickBox();
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				
				if(oNode.parentNode != this)
				{
					this.selectNode(oNode.parentNode);
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				
				return true;
			}
		}
	}
	
}(jaswin.treeview = jaswin.treeview || {}));
